from typing import Optional

import torch
from torch import nn as nn
from torch.nn.init import xavier_uniform_
import torch.nn.functional as F

from ..util.constants import DEFAULT_FEATURES
from ..util.util import mlp


class EARLPerceiverBlock(nn.Module):
    def __init__(self, n_dims, n_heads, activation=F.relu, dim_feedforward=None, concatenate=False, use_norm=True):
        super().__init__()
        if dim_feedforward is None:
            dim_feedforward = 4 * n_dims
        self.concatenate = concatenate

        # Layers
        self.attention = nn.MultiheadAttention(n_dims, n_heads, batch_first=True)
        self.linear1 = nn.Linear((1 + self.concatenate) * n_dims, dim_feedforward)
        self.linear2 = nn.Linear(dim_feedforward, n_dims)

        self.activation = activation

        self.use_norm = use_norm
        if use_norm:
            self.norm1 = nn.LayerNorm(n_dims)
            self.norm2 = nn.LayerNorm(n_dims)
            self.norm3 = nn.LayerNorm(n_dims)
        else:
            self.norm1 = nn.Identity()
            self.norm2 = nn.Identity()
            self.norm3 = nn.Identity()

        self._reset_parameters()

    def _reset_parameters(self):
        r"""Initiate parameters in the transformer model. Taken from PyTorch Transformer impl"""
        for p in self.parameters():
            if p.dim() > 1:
                xavier_uniform_(p)

    def forward(self, src, invariant, mask=None):
        src2 = self.norm1(src)
        invariant = self.norm2(invariant)
        src2, weights = self.attention(src2, invariant, invariant, key_padding_mask=mask)

        src = torch.cat((src, src2), dim=-1) if self.concatenate else (src + src2)

        src2 = self.norm3(src)
        src2 = self.linear2(self.activation(self.linear1(src2)))

        # src = torch.cat((src, src2), dim=-1) if self.concatenate else (src + src2)
        src = src + src2

        return src, weights


class EARLPerceiver(nn.Module):
    def __init__(
            self,
            n_dims: int = 256,
            n_layers: int = 2,
            n_heads: int = 4,
            n_preprocess_layers: int = 1,
            n_postprocess_layers: int = 0,
            query_features: Optional[int] = None,
            key_value_features: Optional[int] = None,
            return_weights=False,
            activation=F.relu
    ):
        """
        EARLPerceiver is an alternative to EARL that uses only a set number of embedding that attend to all the inputs.
        This reduces complexity from O(n^2) to O(n) and gives improved performance specifically on CPU.

        :param n_dims: number of dimensions in the intermediate and output representations.
        :param n_layers: number of encoder layers.
        :param n_heads: number of heads in encoder layers.
        :param n_preprocess_layers: number of dense layers before encoder.
        :param n_postprocess_layers: number of dense layers after encoder.
        :param query_features: number of features in the query input (last dimension).
        :param key_value_features: number of features in the key_value input (last dimension).
        """
        super().__init__()
        if query_features is None:
            query_features = len(DEFAULT_FEATURES)
        if key_value_features is None:
            key_value_features = len(DEFAULT_FEATURES)

        self.n_dims = n_dims
        self.n_layers = n_layers
        self.n_heads = n_heads
        self.return_weights = return_weights

        if n_preprocess_layers > 0:
            self.query_preprocess = mlp(query_features, n_dims, n_preprocess_layers)
            self.key_value_preprocess = mlp(key_value_features, n_dims, n_preprocess_layers)
        else:
            self.query_preprocess = nn.Identity()
            self.key_value_preprocess = nn.Identity()

        self.blocks = nn.ModuleList([
            EARLPerceiverBlock(n_dims, n_heads, activation=activation)
            for _ in range(n_layers)
        ])

        if n_postprocess_layers > 0:
            self.postprocess = mlp(n_dims, n_dims, n_postprocess_layers - 1, n_dims)
        else:
            self.postprocess = nn.Identity()

    def forward(self, query_entities: torch.Tensor, key_value_entities: torch.Tensor,
                mask: Optional[torch.Tensor] = None):
        q_emb = self.query_preprocess(query_entities)
        kv_emb = self.key_value_preprocess(key_value_entities)

        weights = []
        for block in self.blocks:
            q_emb, w = block(q_emb, kv_emb, mask=mask)
            weights.append(w)

        q_emb = self.postprocess(q_emb)

        if self.return_weights:
            return q_emb, tuple(weights)
        return q_emb



import numpy as np
from all.agents import PPO
from all.bodies import Body


class RLBody(Body):
    def __init__(self, agent, n=3):
        super().__init__(agent)
        self.n = n

    def process_action(self, action):
        actions = np.copy(action)
        actions[..., :5] = 2 * actions[..., :5] / (self.n - 1) - 1
        return actions



import numpy as np
from typing import Any, List
from rlgym.utils import common_values
from rlgym.utils.gamestates import PlayerData, GameState, PhysicsObject
from rlgym.utils.obs_builders import ObsBuilder


class AdvancedStacker(ObsBuilder):
    """
    Alternative observation to AdvancedObs. action_stacks past stack_size actions and appends them to AdvancedObs. If there were no previous actions, zeros are assumed as previous actions.

    :param stack_size: number of frames to stack
    """

    def __init__(self, stack_size: int = 15):
        super().__init__()
        self.POS_STD = 6000
        self.VEL_STD = 3000
        self.ANG_STD = 5.5
        self.default_action = np.zeros(common_values.NUM_ACTIONS)
        self.stack_size = stack_size
        self.action_stacks = {}
        self.action_size = self.default_action.shape[0]

    def blank_stack(self, car_id: int) -> None:

    def add_action_to_stack(self, new_action: np.ndarray, car_id: int):
        stack = self.action_stacks[car_id]
        stack[self.action_size:] = stack[:-self.action_size]
        stack[:self.action_size] = new_action

    def reset(self, initial_state: GameState):
        self.action_stacks = {}
        for p in initial_state.players:
            self.action_stacks[p.car_id] = np.concatenate([default_action] * stack_size
            

    def build_obs(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> Any:
        self.add_action_to_stack(previous_action, player.car_id)

        if player.team_num == common_values.ORANGE_TEAM:
            inverted = True
            ball = state.inverted_ball
            pads = state.inverted_boost_pads
        else:
            inverted = False
            ball = state.ball
            pads = state.boost_pads

        obs = [
            ball.position / self.POS_STD,
            ball.linear_velocity / self.VEL_STD,
            ball.angular_velocity / self.ANG_STD,
            previous_action,
            pads,
        ]

        obs.extend(list(self.action_stacks[player.car_id]))

        player_car = self._add_player_to_obs(obs, player, ball, inverted)

        allies = []
        enemies = []

        for other in state.players:
            if other.car_id == player.car_id:
                continue

            if other.team_num == player.team_num:
                team_obs = allies
            else:
                team_obs = enemies

            other_car = self._add_player_to_obs(team_obs, other, ball, inverted)

            # Extra info
            team_obs.extend(
                [
                    (other_car.position - player_car.position) / self.POS_STD,
                    (other_car.linear_velocity - player_car.linear_velocity)
                    / self.VEL_STD,
                ]
            )

        obs.extend(allies)
        obs.extend(enemies)
        return np.concatenate(obs)

    def _add_player_to_obs(self, obs: List, player: PlayerData, ball: PhysicsObject, inverted: bool):
        if inverted:
            player_car = player.inverted_car_data
        else:
            player_car = player.car_data

        rel_pos = ball.position - player_car.position
        rel_vel = ball.linear_velocity - player_car.linear_velocity

        obs.extend(
            [
                rel_pos / self.POS_STD,
                rel_vel / self.VEL_STD,
                player_car.position / self.POS_STD,
                player_car.forward(),
                player_car.up(),
                player_car.linear_velocity / self.VEL_STD,
                player_car.angular_velocity / self.ANG_STD,
                [
                    player.boost_amount,
                    int(player.on_ground),
                    int(player.has_flip),
                    int(player.is_demoed),
                ],
            ]
        )

        return player_car



from typing import Optional, Union, List, Tuple, Any

import numpy as np
from gym import Env, Space
from gym.core import RenderFrame, ActType, ObsType
from gym.spaces import Box
from gym.vector import VectorEnv
from rlgym.envs import Match
from rlgym.gym import Gym
from rlgym.utils import ObsBuilder
from rlgym.utils.action_parsers import DefaultAction, ContinuousAction
from rlgym.utils.gamestates import PlayerData, GameState
from rlgym.utils.reward_functions.common_rewards import ConstantReward


class SimpleObs(ObsBuilder):
    def __init__(self, tick_skip):
        super().__init__()
        self.out = None
        self.tick_skip = tick_skip
        self.boost_pad_seconds_left = np.zeros((34,))

    def get_obs_space(self) -> Space:
        return Box(-np.inf, np.inf, (111,))

    def reset(self, initial_state: GameState):
        self.boost_pad_seconds_left = np.zeros((len(initial_state.boost_pads),))

    def pre_step(self, state: GameState):
        self.out = np.zeros((2 + len(state.boost_pads) + 9 +,))

        self.out[0] = state.blue_score
        self.out[1] = state.orange_score
        self.out[2:2 + len(state.boost_pads)] = self.boost_pad_seconds_left
        s = 2 + len(state.boost_pads)
        self.out[s:s + 3] = state.ball.position
        self.out[s + 3:s + 6] = state.ball.linear_velocity
        self.out[s + 6:s + 9] = state.ball.angular_velocity

        s += 9
        for player in state.players:
            self.out[s] = player.car_id
            self.out[s + 1] = player.team_num
            self.out[s + 2] = player.match_goals
            self.out[s + 3] = player.match_saves
            self.out[s + 4] = player.match_shots
            self.out[s + 5] = player.match_demolishes
            self.out[s + 6] = player.boost_pickups
            self.out[s + 7] = player.is_demoed
            self.out[s + 8] = player.on_ground
            self.out[s + 9] = player.ball_touched
            self.out[s + 10] = player.has_jump
            self.out[s + 11] = player.has_flip
            self.out[s + 12] = player.boost_amount

            self.out[s + 12: s + 15] = player.car_data.position
            self.out[s + 15: s + 18] = player.car_data.linear_velocity
            self.out[s + 18: s + 21] = player.car_data.angular_velocity
            self.out[s + 21: s + 24] = player.car_data.forward()
            self.out[s + 24: s + 27] = player.car_data.up()
            s += 27

        self.boost_pad_seconds_left = (self.boost_pad_seconds_left - self.tick_skip / 120)

    def build_obs(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> Any:
        return self.out


class BetterGym(VectorEnv, Gym):
    def __init__(self, tick_skip=6):
        match = Match(
            reward_function=ConstantReward(),
            terminal_conditions=[],
            obs_builder=SimpleObs(tick_skip),
            action_parser=ContinuousAction()
        )
        super(BetterGym, self).__init__()

    def reset(self, *, seed: Optional[int] = None, options: Optional[dict] = None) -> Tuple[ObsType, dict]:
        pass

    def reset_wait(self, seed: Optional[Union[int, List[int]]] = None, options: Optional[dict] = None):
        pass

    def step(self, action: ActType) -> Tuple[ObsType, float, bool, bool, dict]:
        pass

    def render(self) -> Optional[Union[RenderFrame, List[RenderFrame]]]:
        pass



import numpy as np
from rlgym.utils import RewardFunction
from rlgym.utils.common_values import BLUE_TEAM
from rlgym.utils.gamestates import GameState, PlayerData


class DistributeRewards(RewardFunction):
    """
    Inspired by OpenAI's Dota bot (OpenAI Five).
    Modifies rewards using the formula (1-team_spirit) * own_reward + team_spirit * avg_team_reward - avg_opp_reward
        For instance, in a 3v3 where scoring a goal gives 100 reward with team_spirit 0.3 / 0.6 / 0.9:
            - Goal scorer gets 80 / 60 / 40
            - Teammates get 10 / 20 / 30 each
            - Opponents get -33.3 each

    Note that this will bring mean reward close to zero, so tracking might be misleading.
    If using one of the SB3 envs SB3DistributeRewardsWrapper can be used after logging.
    """
    def __init__(self, reward_func: RewardFunction, team_spirit=0.3):
        super().__init__()
        self.reward_func = reward_func
        self.team_spirit = team_spirit
        self.last_state = None
        self.base_rewards = {}
        self.avg_blue = 0
        self.avg_orange = 0

    def _compute(self, state: GameState, final=False):
        if state != self.last_state:
            self.base_rewards = {}
            sum_blue = 0
            n_blue = 0
            sum_orange = 0
            n_orange = 0
            for player in state.players:
                if final:
                    rew = self.reward_func.get_final_reward(player, state, None)
                else:
                    rew = self.reward_func.get_reward(player, state, None)

                self.base_rewards[player.car_id] = rew
                if player.team_num == BLUE_TEAM:
                    sum_blue += rew
                    n_blue += 1
                else:
                    sum_orange += rew
                    n_orange += 1
            self.avg_blue = sum_blue / (n_blue or 1)
            self.avg_orange = sum_orange / (n_orange or 1)

            self.last_state = state

    def _get_individual_reward(self, player):
        base_reward = self.base_rewards[player.car_id]
        if player.team_num == BLUE_TEAM:
            reward = self.team_spirit * self.avg_blue + (1 - self.team_spirit) * base_reward - self.avg_orange
        else:
            reward = self.team_spirit * self.avg_orange + (1 - self.team_spirit) * base_reward - self.avg_blue
        return reward

    def reset(self, initial_state: GameState):
        self.reward_func.reset(initial_state)

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        self._compute(state, final=False)
        return self._get_individual_reward(player)

    def get_final_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        self._compute(state, final=True)
        return self._get_individual_reward(player)



import math
from typing import Union

import numpy as np

from rlgym.utils import RewardFunction
from rlgym.utils.gamestates import PlayerData, GameState
from rlgym.utils.reward_functions.common_rewards import ConstantReward


class _DummyReward(RewardFunction):
    def reset(self, initial_state: GameState): pass

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float: return 0


class AnnealRewards(RewardFunction):
    """
    Smoothly transitions between reward functions sequentially.

    Example:
        AnnealRewards(rew1, 10_000, rew2, 100_000, rew3)
        will start by only rewarding rew1, then linearly transitions until 10_000 steps, where only rew2 is counted.
        It then transitions between rew2 and rew3, and after 100_000 total steps and further, only rew3 is rewarded.
    """
    STEP = 0
    TOUCH = 1
    GOAL = 2

    def __init__(self, *alternating_rewards_steps: Union[RewardFunction, int],
                 mode: int = STEP, initial_count: int = 0):
        """

        :param alternating_rewards_steps: an alternating sequence of (RewardFunction, int, RewardFunction, int, ...)
                                          specifying reward functions, and the steps at which to transition.
        :param mode: specifies whether to increment counter on steps, touches or goals.
        :param initial_count: the count to start reward calculations at.
        """
        self.rewards_steps = list(alternating_rewards_steps) + [float("inf"), _DummyReward()]
        assert mode in (self.STEP, self.TOUCH, self.GOAL)
        self.mode = mode

        self.last_goals = 0
        self.current_goals = 0

        self.last_transition_step = 0
        self.last_reward = self.rewards_steps.pop(0)
        self.next_transition_step = self.rewards_steps.pop(0)
        self.next_reward = self.rewards_steps.pop(0)
        self.count = initial_count
        self.last_state = None

    def reset(self, initial_state: GameState):
        self.last_reward.reset(initial_state)
        self.next_reward.reset(initial_state)
        self.last_state = None
        while self.next_transition_step < self.count:  # If initial_count is set, find the right rewards
            self._transition(initial_state)

    def _transition(self, state):
        self.last_transition_step = self.next_transition_step
        self.last_reward = self.next_reward
        self.next_transition_step = self.rewards_steps.pop(0)
        self.next_reward = self.rewards_steps.pop(0)
        self.next_reward.reset(state)  # Make sure initial values are set

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        args = player, state, previous_action
        if math.isinf(self.next_transition_step):
            return self.last_reward.get_reward(*args)

        if state != self.last_state:
            if self.mode == self.STEP:
                self.count += 1
            elif self.mode == self.TOUCH and player.ball_touched:
                self.count += 1
            elif self.mode == self.GOAL:
                self.last_goals = self.current_goals
                self.current_goals = state.blue_score + state.orange_score
                if self.current_goals > self.last_goals:
                    self.count += 1
            self.last_state = state

        frac = (self.count - self.last_transition_step) / (self.next_transition_step - self.last_transition_step)
        rew = frac * self.next_reward.get_reward(*args) + (1 - frac) * self.last_reward.get_reward(*args)

        if self.count >= self.next_transition_step:
            self._transition(state)

        return rew



from rlgym.utils import RewardFunction
from rlgym.utils.gamestates import PlayerData, GameState

class JumpTouchReward(RewardFunction):
    """
    a ball touch reward that only triggers when the agent's wheels aren't in contact with the floor
    adjust minimum ball height required for reward with 'min_height' as well as reward scaling with 'exp'
    """

    def __init__(self, min_height=92, exp=0.2):
        self.min_height = min_height
        self.exp = exp

    def reset(self, initial_state: GameState):
        pass

    def get_reward(
            self, player: PlayerData, state: GameState, previous_action: np.ndarray
            ) -> float:
        if player.ball_touched and not player.on_ground and state.ball.position[2] >= self.min_height:
            return ((state.ball.position[2] - 92) ** self.exp)-1

        return 0



from rlgym.utils.reward_functions import RewardFunction

from rlgym.utils.gamestates import GameState, PlayerData

from numpy import ndarray

class SequentialRewards(RewardFunction):
    """ 
    A simple reward class that allows you to transition from one reward to the next at set intervals.
    Example: rewards, step_requirements = [reward1, reward2, reward3, etc], [10_000_000, 20_000_000, 30_000_000, etc]
    my_rewards = SequentialRewards(rewards, step_requirements)   
    """
    def __init__(self, rewards: list, steps: list):
        super().__init__()
        self.rewards_list = rewards
        self.step_counts = steps
        self.step_count = 0
        self.step_index = 0
        assert len(self.rewards_list) == len(self.step_counts)

    def reset(self, initial_state: GameState):
        for rew in self.rewards_list:
            rew.reset(initial_state)

    def get_reward(self, player: PlayerData, state: GameState, previous_action: ndarray) -> float:
        if self.step_index < len(self.step_counts) and self.step_count > self.step_counts[self.step_index]:
            self.step_index += 1

        self.step_count += 1
        return self.rewards_list[self.step_index].get_reward(player, state, previous_action)



import numpy as np

DEFAULT_FEATURES = (
    CLS,
    IS_BALL, IS_BOOST, IS_BLUE, IS_ORANGE,
    POS_X, POS_Y, POS_Z,
    FORWARD_X, FORWARD_Y, FORWARD_Z,
    UP_X, UP_Y, UP_Z,
    VEL_X, VEL_Y, VEL_Z,
    ANG_VEL_X, ANG_VEL_Y, ANG_VEL_Z,
    BOOST_AMOUNT, IS_DEMOED, ON_GROUND, HAS_FLIP
) = range(24)

DEFAULT_FEATURES_STR = (
    "cls",
    "is_ball", "is_boost", "is_blue", "is_orange",
    "pos_x", "pos_y", "pos_z",
    "forward_x", "forward_y", "forward_z",
    "up_x", "up_y", "up_z",
    "vel_x", "vel_y", "vel_z",
    "ang_vel_x", "ang_vel_y", "ang_vel_z",
    "boost_amount", "is_demoed",
    "on_ground", "has_flip"
)

DEFAULT_LABELS = (
    SCORE, NEXT_TOUCH, BOOST_COLLECT, DEMO,
    THROTTLE, STEER, PITCH, YAW, ROLL, JUMP, BOOST, HANDBRAKE,
    RANK
) = range(13)

DEFAULT_LABELS_STR = (
    "score", "next_touch", "boost_collect", "demo",
    "throttle", "steer", "pitch", "yaw", "roll", "jump", "boost", "handbrake",
    "rank"
)

BALL_COLS = (POS_X, POS_Y, POS_Z, VEL_X, VEL_Y, VEL_Z, ANG_VEL_X, ANG_VEL_Y, ANG_VEL_Z)
BOOST_COLS = (POS_X, POS_Y, POS_Z, BOOST_AMOUNT, IS_DEMOED)
PLAYER_COLS = (POS_X, POS_Y, POS_Z,
               FORWARD_X, FORWARD_Y, FORWARD_Z,
               UP_X, UP_Y, UP_Z,
               VEL_X, VEL_Y, VEL_Z,
               ANG_VEL_X, ANG_VEL_Y, ANG_VEL_Z,
               BOOST_AMOUNT, IS_DEMOED, ON_GROUND, HAS_FLIP)

NORM_TRANSFORM = np.array([
    1.,
    1., 1., 1., 1.,
    2300., 2300., 2300.,
    1., 1., 1.,
    1., 1., 1.,
    2300., 2300., 2300.,
    5.5, 5.5, 5.5,
    100., 1., 1., 1.
])  # x_norm = x / NORM_TRANSFORM

SWAP_TRANSFORM = np.array([
    1.,
    1., 1., 1., 1.,
    -1., -1., 1.,
    -1., -1., 1.,
    -1., -1., 1.,
    -1., -1., 1.,
    -1., -1., 1.,
    1., 1., 1., 1.
])

MIRROR_TRANSFORM = np.array([
    1.,
    1., 1., 1., 1.,
    -1., 1., 1.,
    -1., 1., 1.,
    -1., 1., 1.,
    -1., 1., 1.,
    -1., 1., 1.,
    1., 1., 1., 1.
])



from typing import Any

import numpy as np

from rlgym.utils.common_values import NUM_ACTIONS
from rlgym.utils.gamestates import PlayerData, GameState
from rlgym.utils.obs_builders import ObsBuilder


class GeneralStacker(ObsBuilder):
    """
    Stacks several observations into one

    :param obs: observation that is to be stacked.
    :param stack_size: how many past observations the queue hold.

    """

    def __init__(self, obs: ObsBuilder, stack_size: int = 15):
        super().__init__()
        self.stack_size = stack_size
        self.obs_builder = obs

    def reset(self, initial_state: GameState):
        self.obs = {}
        previous_action = np.zeros(NUM_ACTIONS)
        for player in initial_state.players:
            self.obs_builder.reset(initial_state=initial_state)
            initial_obs = self.obs_builder.build_obs(player, initial_state, previous_action=previous_action)
            self.obs[player.car_id] = np.concatenate([initial_obs] * self.stack_size)
        self.obs_builder.reset(initial_state=initial_state)

    def build_obs(
            self, player: PlayerData, state: GameState, previous_action: np.ndarray
    ) -> Any:
        frame = self.obs_builder.build_obs(player, state, previous_action)

        num = frame.shape[0]
        obs = self.obs[player.car_id]
        obs[num:] = obs[:-num]
        obs[:num] = frame

        return obs




from rlgym.utils.state_setters import StateSetter
from rlgym.utils.state_setters import StateWrapper
from rlgym.utils.common_values import BALL_RADIUS, CEILING_Z, BLUE_TEAM, ORANGE_TEAM
import numpy as np
from numpy import random as rand

X_MAX = 7000
Y_MAX = 9000
Z_MAX_CAR = 1900
GOAL_HEIGHT = 642.775
PITCH_MAX = np.pi / 2
ROLL_MAX = np.pi
BLUE_GOAL_POSITION = np.asarray([0, -5120, 0])
ORANGE_GOAL_POSITION = np.asarray([0, 5120, 0])


class HoopsLikeSetter(StateSetter):

    def __init__(self, spawn_radius: float = 800):
        """
        Hoops-like kickoff constructor

        :param spawn_radius: Float determining how far away to spawn the cars from the ball.
        """
        super().__init__()
        self.spawn_radius = spawn_radius
        self.yaw_vector = np.asarray([-1, 0, 0])

    def reset(self, state_wrapper: StateWrapper):
        """
        Modifies the StateWrapper to contain random values the ball and each car.

        :param state_wrapper: StateWrapper object to be modified with desired state values.
        """
        self._reset_ball_random(state_wrapper)
        self._reset_cars_random(state_wrapper, self.spawn_radius)

    def _reset_ball_random(self, state_wrapper: StateWrapper):
        """
        Resets the ball according to a uniform distribution along y-axis and normal distribution along x-axis.

        :param state_wrapper: StateWrapper object to be modified.
        """
        new_x = rand.uniform(-4000, 4000)
        new_y = rand.triangular(-5000, 0, 5000)
        new_z = rand.uniform(GOAL_HEIGHT, CEILING_Z - 2 * BALL_RADIUS)
        state_wrapper.ball.set_pos(new_x, new_y, new_z)
        state_wrapper.ball.set_ang_vel(0, 0, 0)
        state_wrapper.ball.set_lin_vel(0, 0, 0)

    def _reset_cars_random(self, state_wrapper: StateWrapper, spawn_radius: float):
        """
        Function to set all cars inbetween the ball and net roughly facing the ball. The other cars will be spawned
        randomly.

        :param state_wrapper: StateWrapper object to be modified.
        :param spawn_radius: Float determining how far away to spawn the cars from the ball.
        """
        orange_done = False
        blue_done = False
        for i, car in enumerate(state_wrapper.cars):
            if car.team_num == BLUE_TEAM and not blue_done:
                # just shorthands for ball_x and ball_y
                bx = state_wrapper.ball.position[0]
                by = state_wrapper.ball.position[1]

                # add small variation to spawn radius
                car_spawn_radius = rand.triangular(0.8, 1, 1.2) * spawn_radius

                # calculate distance from ball to goal.
                R = ((BLUE_GOAL_POSITION[0] - bx) ** 2 + (BLUE_GOAL_POSITION[1] - by) ** 2) ** 0.5

                # use similarity of triangles to calculate offsets
                x_offset = ((BLUE_GOAL_POSITION[0] - bx) / R) * car_spawn_radius
                y_offset = ((BLUE_GOAL_POSITION[1] - by) / R) * car_spawn_radius

                # offset the car's positions
                car.set_pos(bx + x_offset, by + y_offset, 17)

                # compute vector from ball to car
                rel_ball_car_vector = car.position - state_wrapper.ball.position

                # calculate the angle between the yaw vector and relative vector and use that as yaw.

                yaw = np.arccos(np.dot(rel_ball_car_vector / np.linalg.norm(rel_ball_car_vector), self.yaw_vector))

                # then sprinkle in more variation by offsetting by random angle up to pi/8 radians (this is arbitrary)
                max_offset_angle = np.pi / 8
                yaw += rand.random() * max_offset_angle * 2 - max_offset_angle

                # random at least slightly above 0 boost amounts
                car.boost = rand.uniform(0.2, 1)

                # make car face the ball roughly
                car.set_rot(pitch=0, roll=0, yaw=yaw)

                # x angular velocity (affects pitch) set to 0
                # y angular velocity (affects roll) set to 0
                # z angular velocity (affects yaw) set to 0
                car.set_ang_vel(x=0, y=0, z=0)
                car.set_lin_vel(x=0, y=0, z=0)
                blue_done = True
            elif car.team_num == ORANGE_TEAM and not orange_done:
                # just shorthands for ball_x and ball_y
                bx = state_wrapper.ball.position[0]
                by = state_wrapper.ball.position[1]

                # add small variation to spawn radius
                car_spawn_radius = rand.triangular(0.8, 1, 1.2) * spawn_radius

                # calculate distance from ball to goal.
                R = ((ORANGE_GOAL_POSITION[0] - bx) ** 2 + (ORANGE_GOAL_POSITION[1] - by) ** 2) ** 0.5

                # use similarity of triangles to calculate offsets
                x_offset = ((ORANGE_GOAL_POSITION[0] - bx) / R) * car_spawn_radius
                y_offset = ((ORANGE_GOAL_POSITION[1] - by) / R) * car_spawn_radius

                # offset the car's positions
                car.set_pos(bx + x_offset, by + y_offset, 17)

                # compute vector from ball to car
                rel_ball_car_vector = car.position - state_wrapper.ball.position

                # calculate the angle between the yaw vector and relative vector and use that as yaw.

                yaw = np.arccos(np.dot(rel_ball_car_vector / np.linalg.norm(rel_ball_car_vector), self.yaw_vector))

                # then sprinkle in more variation by offsetting by random angle up to pi/8 radians (this is arbitrary)
                max_offset_angle = np.pi / 8
                yaw += rand.random() * max_offset_angle * 2 - max_offset_angle

                # random at least slightly above 0 boost amounts
                car.boost = rand.uniform(0.2, 1)

                # make car face the ball roughly
                car.set_rot(pitch=0, roll=0, yaw=yaw - np.pi if yaw > 0 else yaw + np.pi)

                # x angular velocity (affects pitch) set to 0
                # y angular velocity (affects roll) set to 0
                # z angular velocity (affects yaw) set to 0
                car.set_ang_vel(x=0, y=0, z=0)
                car.set_lin_vel(x=0, y=0, z=0)
                orange_done = True
            else:
                car.set_pos(rand.uniform(-4096,4096)),rand.uniform(0,-1**(car.team_num-1)*5120,17)

                car.boost = rand.uniform(0.2, 1)

                # make car face the ball roughly
                car.set_rot(pitch=0, roll=0, yaw=rand.uniform(-np.pi, np.pi))

                # x angular velocity (affects pitch) set to 0
                # y angular velocity (affects roll) set to 0
                # z angular velocity (affects yaw) set to 0
                car.set_ang_vel(x=0, y=0, z=0)
                car.set_lin_vel(x=0, y=0, z=0)



from rlgym.utils import RewardFunction
from rlgym.utils.state_setters import StateSetter
from rlgym.utils.state_setters import StateWrapper
from random import choices
from typing import Sequence, Union, Tuple


class WeightedSampleSetter(StateSetter):
    """
    Samples StateSetters randomly according to their weights.

    :param state_setters: 1-D array-like of state-setters to be sampled from
    :param weights: 1-D array-like of the weights associated with each entry in state_setters
    """

    def __init__(self, state_setters: Sequence[StateSetter], weights: Sequence[float]):
        super().__init__()
        self.state_setters = state_setters
        self.weights = weights
        assert len(state_setters) == len(weights), \
            f"Length of state_setters should match the length of weights, " \
            f"instead lengths {len(state_setters)} and {len(weights)} were given respectively."

    @classmethod
    def from_zipped(
            cls,
            *setters_and_weights: Union[StateSetter, Tuple[RewardFunction, float]]
    ) -> "WeightedSampleSetter":
        """
        Alternate constructor which takes any number of either rewards, or (reward, weight) tuples.
        :param setters_and_weights: a sequence of RewardFunction or (RewardFunction, weight) tuples
        """
        rewards = []
        weights = []
        for value in setters_and_weights:
            if isinstance(value, tuple):
                r, w = value
            else:
                r, w = value, 1.
            rewards.append(r)
            weights.append(w)
        return cls(tuple(rewards), tuple(weights))

    def reset(self, state_wrapper: StateWrapper):
        """
        Executes the reset of randomly sampled state-setter

        :param state_wrapper: StateWrapper object to be modified with desired state values.
        """
        choices(self.state_setters, weights=self.weights)[0].reset(state_wrapper)



from rlgym.utils.state_setters import StateSetter
from rlgym.utils.state_setters import StateWrapper
from rlgym.utils.math import rand_vec3
import numpy as np
from numpy import random as rand
import random

BALL_RADIUS = 94
DEG_TO_RAD = 3.14159265 / 180

class WallPracticeState(StateSetter):

    def __init__(self, air_dribble_odds=1/3, backboard_roll_odds=1/3, side_high_odds=1/3):
        """
        WallPracticeState to setup wall practice
        """
        super().__init__()

        self.air_dribble_odds = air_dribble_odds
        self.backboard_roll_odds = backboard_roll_odds
        self.side_high_odds = side_high_odds

    def reset(self, state_wrapper: StateWrapper):
        choice_list = [0] * int(self.backboard_roll_odds * 100) + \
                      [1] * int(self.side_high_odds * 100) + \
                      [2] * int(self.air_dribble_odds * 100)
        scenario_pick = random.choice(choice_list)

        if scenario_pick == 0:
            self._short_goal_roll(state_wrapper)
        elif scenario_pick == 1:
            self._side_high_roll(state_wrapper)
        elif scenario_pick == 2:
            self._air_dribble_setup(state_wrapper)

    def _air_dribble_setup(self, state_wrapper):
        """
        A medium roll up a side wall with the car facing the roll path

        :param state_wrapper:
        """

        axis_inverter = 1 if random.randrange(2) == 1 else -1
        team_side = 0 if random.randrange(2) == 1 else 1
        team_inverter = 1 if team_side == 0 else -1

        #if only 1 play, team is always 0

        ball_x_pos = 3000 * axis_inverter
        ball_y_pos = random.randrange(7600) - 3800
        ball_z_pos = BALL_RADIUS
        state_wrapper.ball.set_pos(ball_x_pos, ball_y_pos, ball_z_pos)

        ball_x_vel = (2000 + (random.randrange(1000) - 500)) * axis_inverter
        ball_y_vel = random.randrange(1000) * team_inverter
        ball_z_vel = 0
        state_wrapper.ball.set_lin_vel(ball_x_vel, ball_y_vel, ball_z_vel)

        chosen_car = [car for car in state_wrapper.cars if car.team_num == team_side][0]
        #if randomly pick, chosen_car is from orange instead

        car_x_pos = 2500 * axis_inverter
        car_y_pos = ball_y_pos
        car_z_pos = 27

        yaw = 0 if axis_inverter == 1 else 180
        car_pitch_rot = 0 * DEG_TO_RAD
        car_yaw_rot = (yaw + (random.randrange(40) - 20)) * DEG_TO_RAD
        car_roll_rot = 0 * DEG_TO_RAD

        chosen_car.set_pos(car_x_pos, car_y_pos, car_z_pos)
        chosen_car.set_rot(car_pitch_rot, car_yaw_rot, car_roll_rot)
        chosen_car.boost = 100

        for car in state_wrapper.cars:
            if car is chosen_car:
                continue

            # set all other cars randomly in the field
            car.set_pos(random.randrange(2944) - 1472, random.randrange(3968) - 1984, 0)
            car.set_rot(0, (random.randrange(360) - 180) * (3.1415927 / 180), 0)

    def _side_high_roll(self, state_wrapper):
        """
        A high vertical roll up the side of the field

        :param state_wrapper:
        """
        sidepick = random.randrange(2)

        side_inverter = 1
        if sidepick == 1:
            # change side
            side_inverter = -1


        # MAGIC NUMBERS ARE FROM MANUAL CALIBRATION AND WHAT FEELS RIGHT

        ball_x_pos = 3000 * side_inverter
        ball_y_pos = random.randrange(1500) - 750
        ball_z_pos = BALL_RADIUS
        state_wrapper.ball.set_pos(ball_x_pos, ball_y_pos, ball_z_pos)

        ball_x_vel = (2000 + random.randrange(1000) - 500) * side_inverter
        ball_y_vel = random.randrange(1500) - 750
        ball_z_vel = random.randrange(300)
        state_wrapper.ball.set_lin_vel(ball_x_vel, ball_y_vel, ball_z_vel)

        wall_car_blue = [car for car in state_wrapper.cars if car.team_num == 0][0]

        #blue car setup
        blue_pitch_rot = 0 * DEG_TO_RAD
        blue_yaw_rot = 90 * DEG_TO_RAD
        blue_roll_rot = 90 * side_inverter * DEG_TO_RAD
        wall_car_blue.set_rot(blue_pitch_rot, blue_yaw_rot, blue_roll_rot)

        blue_x = 4096 * side_inverter
        blue_y = -2500 + (random.randrange(500) - 250)
        blue_z = 600 + (random.randrange(400) - 200)
        wall_car_blue.set_pos(blue_x, blue_y, blue_z)
        wall_car_blue.boost = 100

        #orange car setup
        wall_car_orange = None
        if len(state_wrapper.cars) > 1:
            wall_car_orange = [car for car in state_wrapper.cars if car.team_num == 1][0]
            # orange car setup
            orange_pitch_rot = 0 * DEG_TO_RAD
            orange_yaw_rot = -90 * DEG_TO_RAD
            orange_roll_rot = -90 * side_inverter * DEG_TO_RAD
            wall_car_orange.set_rot(orange_pitch_rot, orange_yaw_rot, orange_roll_rot)

            orange_x = 4096 * side_inverter
            orange_y = 2500 + (random.randrange(500) - 250)
            orange_z = 400 + (random.randrange(400) - 200)
            wall_car_orange.set_pos(orange_x, orange_y, orange_z)
            wall_car_orange.boost = 100

        for car in state_wrapper.cars:
            if len(state_wrapper.cars) == 1 or car is wall_car_orange or car is wall_car_blue:
                continue

            # set all other cars randomly in the field
            car.set_pos(random.randrange(2944) - 1472, random.randrange(3968) - 1984, 0)
            car.set_rot(0, (random.randrange(360) - 180) * (3.1415927/180), 0)

    def _short_goal_roll(self, state_wrapper):
        """
        A short roll across the backboard and down in front of the goal

        :param state_wrapper:
        :return:
        """

        if len(state_wrapper.cars) > 1:
            defense_team = random.randrange(2)
        else:
            defense_team = 0
        sidepick = random.randrange(2)

        defense_inverter = 1
        if defense_team == 0:
            # change side
            defense_inverter = -1

        side_inverter = 1
        if sidepick == 1:
            # change side
            side_inverter = -1

        # MAGIC NUMBERS ARE FROM MANUAL CALIBRATION AND WHAT FEELS RIGHT

        x_random = random.randrange(446)
        ball_x_pos = (-2850 + x_random) * side_inverter
        ball_y_pos = (5120 - BALL_RADIUS) * defense_inverter
        ball_z_pos = 1400 + random.randrange(400) - 200
        state_wrapper.ball.set_pos(ball_x_pos, ball_y_pos, ball_z_pos)

        ball_x_vel = (1000 + random.randrange(400) - 200) * side_inverter
        ball_y_vel = 0
        ball_z_vel = 550
        state_wrapper.ball.set_lin_vel(ball_x_vel, ball_y_vel, ball_z_vel)


        wall_car = [car for car in state_wrapper.cars if car.team_num == defense_team][0]

        wall_car_x = (2000 - random.randrange(500)) * side_inverter
        wall_car_y = 5120 * defense_inverter
        wall_car_z = 1000 + (random.randrange(500) - 500)
        wall_car.set_pos(wall_car_x, wall_car_y, wall_car_z)

        wall_pitch_rot = (0 if side_inverter == -1 else 180) * DEG_TO_RAD
        wall_yaw_rot = 0 * DEG_TO_RAD
        wall_roll_rot = -90 * defense_inverter * DEG_TO_RAD
        wall_car.set_rot(wall_pitch_rot, wall_yaw_rot, wall_roll_rot)
        wall_car.boost = 25

        if len(state_wrapper.cars) > 1:
            challenge_car = [car for car in state_wrapper.cars if car.team_num != defense_team][0]
            challenge_car.set_pos(0, 1000 * defense_inverter, 0)

            challenge_pitch_rot = 0 * DEG_TO_RAD
            challenge_yaw_rot = 90 * defense_inverter * DEG_TO_RAD
            challenge_roll_rot = 0 * DEG_TO_RAD
            challenge_car.set_rot(challenge_pitch_rot, challenge_yaw_rot, challenge_roll_rot)
            challenge_car.boost = 100

        for car in state_wrapper.cars:
            if len(state_wrapper.cars) == 1 or car is wall_car or car is challenge_car:
                continue

            car.set_pos(random.randrange(2944) - 1472, (-4500 + random.randrange(500) - 250) * defense_inverter, 0)
            car.set_rot(0, (random.randrange(360) - 180) * DEG_TO_RAD, 0)



from ray.rllib.algorithms.apex_dqn import ApexDQNConfig
from test_marl_env import RLLibEnv
from ray.tune import register_env
from rlgym_tools.extra_action_parsers.lookup_act import LookupAction

def make_env():
    import rlgym
    team_size = 1
    self_play = True
    n_agents = team_size*2 if self_play else team_size

    env = rlgym.make(action_parser=LookupAction(),
                     team_size=team_size,
                     spawn_opponents=self_play,
                     use_injector=True)

    return RLLibEnv(env)


if __name__ == "__main__":
    register_env("rlgym", lambda c: make_env())

    cfg = ApexDQNConfig()
    cfg.framework("torch")
    cfg.rollouts(num_rollout_workers=2, )
    cfg.resources(num_gpus=1)
    cfg.environment(env="rlgym")
    cfg.num_atoms = 51

    algo = cfg.build()

    for i in range(100000000):
        result = algo.train()
        print("Epoch: {}\n"
              "Mean Reward: {:7.3f}\n"
              "Mean Ep Len: {:7.2f}\n"
              "Timesteps: {}\n"
              "Time elapsed: {:7.2f}\n"
              .format(result["training_iteration"],
                      result["episode_reward_mean"],
                      result["episode_len_mean"],
                      result["timesteps_total"],
                      result["time_total_s"]))

        if i % 5 == 0:
            checkpoint_dir = algo.save()
            print(f"Checkpoint saved in directory {checkpoint_dir}")



from typing import Any, Tuple, Dict, Union

import numpy as np
from maze.core.env.core_env import CoreEnv
from maze.core.env.environment_context import EnvironmentContext
from maze.core.env.maze_action import MazeActionType
from maze.core.env.maze_env import MazeEnv
from maze.core.env.maze_state import MazeStateType
from maze.core.env.structured_env import StepKeyType, ActorID
from maze.core.rendering.renderer import Renderer

from rlgym.envs import Match
from rlgym.gym import Gym


class MazeRLGymEnv(CoreEnv):
    def __init__(self, env: Gym):
        super().__init__()
        self.env = env
        self.current_state = None
        self.current_actions = None
        self.n_agents = env._match.agents
        self.current_agent = 0

    def step(self, maze_action: MazeActionType) -> \
            Tuple[MazeStateType, Union[float, np.ndarray, Any], bool, Dict[Any, Any]]:

        if self.current_agent == self.n_agents - 1:
            self.env.step(self.current_actions)
            self.current_agent = 0
            EnvironmentContext.increment_env_step()
        else:
            self.current_agent += 1

    def reset(self) -> MazeStateType:
        self.current_actions = []
        self.current_state = []
        return self.env.reset()

    def seed(self, seed: int) -> None:
        return self.env.seed(seed)

    def close(self) -> None:
        return self.env.close()

    def get_maze_state(self) -> MazeStateType:
        pass

    def get_serializable_components(self) -> Dict[str, Any]:
        pass

    def get_renderer(self) -> Renderer:
        raise NotImplementedError

    def actor_id(self) -> ActorID:
        pass

    def is_actor_done(self) -> bool:
        pass

    @property
    def agent_counts_dict(self) -> Dict[StepKeyType, int]:
        pass

    def clone_from(self, env: 'CoreEnv') -> None:
        pass



import gc
import os
import pickle

import numpy as np
import torch
import tqdm
from torch.utils import data as data
from torch.utils.data.dataset import T_co, TensorDataset, ConcatDataset, IterableDataset

from .create_dataset import replay_to_dfs, convert_dfs, normalize, swap_teams, swap_left_right, get_base_features


class ReplayDataset(IterableDataset):
    def __init__(self, replay_folder, cache_folder, limit=np.inf, batch_size=512, buffer_size=262144):
        self.replay_folder = replay_folder
        self.cache_folder = cache_folder
        self.limit = limit
        self.batch_size = batch_size
        self.buffer_size = buffer_size
        self.buffer_x, self.buffer_y = get_base_features(buffer_size, 6)
        self.n = 0

    def __iter__(self):
        worker_info = torch.utils.data.get_worker_info()
        if self.replay_folder is None:
            files = [(dp, f) for dp, dn, fn in os.walk(self.cache_folder) for f in fn if f.endswith(".pickle")]
        else:
            files = [(dp, f) for dp, dn, fn in os.walk(self.replay_folder) for f in fn if f.endswith(".replay")]
        if worker_info is None:
            start = 0
            step = 1
        else:
            start = worker_info.id
            step = worker_info.num_workers

        for dp, f in files[start::step]:
            try:
                if self.replay_folder is None:
                    out_path = os.path.join(dp, f)
                    with open(out_path, "rb") as handle:
                        dfs = pickle.load(handle)
                else:
                    in_path = os.path.join(dp, f)
                    out_path = os.path.join(self.cache_folder, f[:-7] + ".pickle")
                    if os.path.exists(out_path):
                        with open(out_path, "rb") as handle:
                            dfs = pickle.load(handle)
                    else:
                        dfs = replay_to_dfs(in_path)
                        with open(out_path, "wb") as handle:
                            pickle.dump(dfs, handle)

                x_n, y_n = convert_dfs(dfs)
                assert x_n[2].shape == x_n[3].shape
                normalize(x_n)

                x_s, y_s = [np.copy(v) for v in x_n], [np.copy(v) for v in y_n]
                swap_teams(x_n, y_n)

                x_m, y_m = [np.copy(v) for v in x_n], [np.copy(v) for v in y_n]
                swap_left_right(x_m, y_m)

                x_sm, y_sm = [np.copy(v) for v in x_s], [np.copy(v) for v in y_s]
                swap_left_right(x_sm, y_sm)

                added_size = len(x_n[0]) * 4
                if self.n + added_size >= self.buffer_size:
                    indices = np.random.permutation(self.n)
                    self.buffer_x = [v[indices] for v in self.buffer_x]
                    self.buffer_y = [v[indices] for v in self.buffer_y]
                    for i in range(0, self.n, self.batch_size):
                        batch_x = tuple(torch.from_numpy(t[i:i + self.batch_size]).cuda() for t in self.buffer_x)
                        batch_y = tuple(torch.from_numpy(t[i:i + self.batch_size]).cuda() for t in self.buffer_y)
                        yield batch_x, batch_y
                    self.buffer_x, self.buffer_y = get_base_features(self.buffer_size, 6)
                    self.n = 0

                self._fill_buffer(self.buffer_x, self.n, x_n, x_s, x_m, x_sm)
                self._fill_buffer(self.buffer_y, self.n, y_n, y_s, y_m, y_sm)

                self.n += added_size

            except Exception as e:
                # print(e)
                pass

    def _fill_buffer(self, start, buffer, *arrs):
        for ds, v, v_s, v_m, v_sm in zip(buffer, *arrs):
            j = start
            ds[j:j + len(v)] = v
            j += len(v)
            ds[j:j + len(v)] = v_s
            j += len(v_s)
            ds[j:j + len(v)] = v_m
            j += len(v_m)
            ds[j:j + len(v)] = v_sm
            j += len(v_sm)

    def __len__(self):
        return self.limit


class ReplayDatasetFull(data.Dataset):
    def __init__(self, replay_folder, cache_folder, limit=None, name=None):
        self.replay_folder = replay_folder
        self.cache_folder = cache_folder
        self.limit = limit

        if name is not None:
            save_path = os.path.join(cache_folder, f"{name}.npz")
            if os.path.exists(save_path):
                arrs = np.load(save_path)
                arrs = list(arrs.values())
                self.x, self.y = arrs[:4], arrs[4:]
                swap_teams(self.x, self.y, slice(None, None, 2))
                return

        if self.replay_folder is None:
            files = [(dp, f) for dp, dn, fn in os.walk(self.cache_folder) for f in fn if f.endswith(".pickle")]
        else:
            files = [(dp, f) for dp, dn, fn in os.walk(self.replay_folder) for f in fn if f.endswith(".replay")]

        file_iter = tqdm.tqdm(enumerate(files[:limit]),
                              desc="Load",
                              total=limit,
                              bar_format="{l_bar}{r_bar}")

        arrays = []
        for i, (dp, f) in file_iter:
            try:
                if self.replay_folder is None:
                    out_path = os.path.join(dp, f)
                    with open(out_path, "rb") as handle:
                        dfs = pickle.load(handle)
                else:
                    in_path = os.path.join(dp, f)
                    out_path = os.path.join(self.cache_folder, f[:-7] + ".pickle")
                    if os.path.exists(out_path):
                        with open(out_path, "rb") as handle:
                            dfs = pickle.load(handle)
                    else:
                        dfs = replay_to_dfs(in_path)
                        with open(out_path, "wb") as handle:
                            pickle.dump(dfs, handle)

                x_n, y_n = convert_dfs(dfs)
                assert x_n[2].shape == x_n[3].shape
                normalize(x_n)

                arrays.append((x_n, y_n))

                # x_s, y_s = [v.copy() for v in x_n], [v.copy() for v in y_n]
                # swap_teams(x_s, y_s)

                # arrays.append((x_s, y_s))
            except Exception as e:
                print(e)
                pass

        print("Concatenating...")
        self.x = []
        for i in range(len(arrays[0][0])):
            conc = []
            for row in arrays:
                conc.append(row[0][i])
                row[0][i] = None
            self.x.append(np.concatenate(conc))
            conc.clear()
            gc.collect()

        self.y = []
        for i in range(len(arrays[0][1])):
            conc = []
            for row in arrays:
                conc.append(row[1][i])
                row[1][i] = None
            self.y.append(np.concatenate(conc))
            conc.clear()
            gc.collect()

        if name is not None:
            np.savez_compressed(save_path, *self.x, *self.y)
        swap_teams(self.x, self.y, slice(None, None, 2))

    def __getitem__(self, index) -> T_co:
        return [v[index] for v in self.x], [v[index] for v in self.y]

    def __len__(self):
        return self.x[0].shape[0]


def get_dataset(replay_folder, cache_folder, limit, name=None):
    if name is not None:
        name = os.path.join(cache_folder, name)
        if os.path.exists(name):
            return torch.load(name)

    if replay_folder is None:
        files = [(dp, f) for dp, dn, fn in os.walk(cache_folder) for f in fn if f.endswith(".pickle")]
    else:
        files = [(dp, f) for dp, dn, fn in os.walk(replay_folder) for f in fn if f.endswith(".replay")]

    file_iter = tqdm.tqdm(enumerate(files[:limit]),
                          desc="Load",
                          total=limit,
                          bar_format="{l_bar}{r_bar}")

    datasets = []
    for i, (dp, f) in file_iter:
        try:
            if replay_folder is None:
                out_path = os.path.join(dp, f)
                with open(out_path, "rb") as handle:
                    dfs = pickle.load(handle)
            else:
                in_path = os.path.join(dp, f)
                out_path = os.path.join(cache_folder, f[:-7] + ".pickle")
                if os.path.exists(out_path):
                    with open(out_path, "rb") as handle:
                        dfs = pickle.load(handle)
                else:
                    dfs = replay_to_dfs(in_path)
                    with open(out_path, "wb") as handle:
                        pickle.dump(dfs, handle)

            x_n, y_n = convert_dfs(dfs, tensors=True)
            assert x_n[2].shape == x_n[3].shape
            normalize(x_n)

            swap_teams(x_n, y_n, slice(i % 2, None, 2))

            datasets.append(TensorDataset(*x_n, *y_n))

            # x_s, y_s = [v.copy() for v in x_n], [v.copy() for v in y_n]
            # swap_teams(x_s, y_s)

            # arrays.append((x_s, y_s))
        except Exception as e:
            print(e)
            pass

    ds = ConcatDataset(datasets)
    if name is not None:
        torch.save(ds, name)
    return ds



from rlgym.utils.gamestates import PlayerData, GameState, PhysicsObject
from rlgym.utils.obs_builders import ObsBuilder
from typing import Any, List
from rlgym.utils import common_values
import numpy as np
import math


class AdvancedObsPadder(ObsBuilder):
    """adds 0 padding to accommodate differing numbers of agents"""
    def __init__(self, team_size=3, expanding=False):
        super().__init__()
        self.team_size = team_size
        self.POS_STD = 2300
        self.ANG_STD = math.pi
        self.expanding = expanding

    def reset(self, initial_state: GameState):
        pass

    def build_obs(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> Any:

        if player.team_num == common_values.ORANGE_TEAM:
            inverted = True
            ball = state.inverted_ball
            pads = state.inverted_boost_pads
        else:
            inverted = False
            ball = state.ball
            pads = state.boost_pads

        obs = [ball.position / self.POS_STD,
               ball.linear_velocity / self.POS_STD,
               ball.angular_velocity / self.ANG_STD,
               previous_action,
               pads]

        player_car = self._add_player_to_obs(obs, player, ball, inverted)

        allies = []
        enemies = []
        ally_count = 0
        enemy_count = 0
        for other in state.players:
            if other.car_id == player.car_id:
                continue

            if other.team_num == player.team_num:
                team_obs = allies
                ally_count += 1
                if ally_count > self.team_size-1:
                    continue
            else:
                team_obs = enemies
                enemy_count += 1
                if enemy_count > self.team_size:
                    continue
            other_car = self._add_player_to_obs(team_obs, other, ball, inverted)

            # Extra info
            team_obs.extend([
                (other_car.position - player_car.position) / self.POS_STD,
                (other_car.linear_velocity - player_car.linear_velocity) / self.POS_STD
            ])

        while ally_count < self.team_size-1:
            self._add_dummy(allies)
            ally_count += 1

        while enemy_count < self.team_size:
            self._add_dummy(enemies)
            enemy_count += 1

        obs.extend(allies)
        obs.extend(enemies)
        if self.expanding:
            return np.expand_dims(np.concatenate(obs), 0)
        return np.concatenate(obs)

    def _add_dummy(self, obs: List):
        obs.extend([
            np.zeros(3),
            np.zeros(3),
            np.zeros(3),
            np.zeros(3),
            np.zeros(3),
            np.zeros(3),
            np.zeros(3),
            [0, 0, 0, 0, 0]])
        obs.extend([np.zeros(3), np.zeros(3)])

    def _add_player_to_obs(self, obs: List, player: PlayerData, ball: PhysicsObject, inverted: bool):
        if inverted:
            player_car = player.inverted_car_data
        else:
            player_car = player.car_data

        rel_pos = ball.position - player_car.position
        rel_vel = ball.linear_velocity - player_car.linear_velocity

        obs.extend([
            rel_pos / self.POS_STD,
            rel_vel / self.POS_STD,
            player_car.position / self.POS_STD,
            player_car.forward(),
            player_car.up(),
            player_car.linear_velocity / self.POS_STD,
            player_car.angular_velocity / self.ANG_STD,
            [player.boost_amount,
             int(player.on_ground),
             int(player.has_flip),
             int(player.is_demoed),
             int(player.has_jump)]])

        return player_car


if __name__ == "__main__":
    pass



import torch
from torch import nn as nn


class NextGoalPredictor(nn.Linear):
    def __init__(self, n_dims, **kwargs):
        super().__init__(n_dims, 2, **kwargs)


class InteractionPredictor(nn.Module):
    def __init__(self, input_dim, emb_dim):
        # Does a single linear layer and dot product between sequences
        super().__init__()
        self.input_dim = input_dim
        self.emb_dim = emb_dim

        self.linear1 = nn.Linear(input_dim, emb_dim)
        self.linear2 = nn.Linear(input_dim, emb_dim)

    def forward(self, inp1, inp2):
        # Returns a tensor of shape (bs, inp1.shape[1], inp2.shape[1])
        # Can be used for things like touch, demo and boost grabs
        # For instance, for predicting next boost one could return logits of (bs, n_players, n_boosts)
        bs = inp1.shape[0]
        emb1 = self.linear1(inp1)
        emb2 = self.linear2(inp2)
        res = torch.bmm(emb1.view(bs, -1, self.emb_dim), emb2.view(bs, self.emb_dim, -1))
        return res


class ControlsPredictorContinuous(nn.Module):
    def __init__(self, n_dims):
        super().__init__()
        self.linear = nn.Linear(n_dims, 8)

    def forward(self, emb):
        actions = self.linear(emb)
        return torch.clip(actions, -1, 1)


class ControlsPredictorDiscrete(nn.Module):
    DEFAULT_SPLITS = (3, 3, 3, 3, 3, 2, 2, 2)

    def __init__(self, n_dims, splits=None):
        super().__init__()
        if splits is None:
            splits = self.DEFAULT_SPLITS
        self.splits = splits
        self.linear = nn.Linear(n_dims, sum(splits))

    def forward(self, emb):
        actions = self.linear(emb)
        return torch.split(actions, self.splits, dim=-1)



"""
A module containing an accurate instantaneous fps logging callback
"""

from time import time

from stable_baselines3.common.callbacks import BaseCallback


class SB3InstantaneousFPSCallback(BaseCallback):
  
  def __init__(self, average_over_n_rollouts:int=5):
    if average_over_n_rollouts < 1:
      raise ValueError("average_over_n_rollouts must be greater than 0")
    super().__init__()
    self.last_steps = []
    self.last_times = []
    self.capacity = average_over_n_rollouts + 1

  def _on_step(self):
    return True

  def _init_callback(self):
    self.last_steps.append(self.model.num_timesteps)
    self.last_times.append(time())

  def _on_rollout_end(self):
    if len(self.last_steps) == self.capacity:
      self.last_steps.pop(0)
      self.last_times.pop(0)
    self.last_steps.append(self.model.num_timesteps)
    self.last_times.append(time())
    
    fps = (self.last_steps[-1] - self.last_steps[0]) / (self.last_times[-1] - self.last_times[0])

    self.model.logger.record('time/instantaneous_fps', fps)
    



from .sb3_distribute_rewards_wrapper import SB3DistributeRewardsWrapper
from .sb3_log_reward import SB3LogReward
from .sb3_multiple_instance_env import SB3MultipleInstanceEnv
from .sb3_single_instance_env import SB3SingleInstanceEnv
from .sb3_multi_agent_tools import multi_learn



from typing import Optional

import torch
from torch import nn as nn, Tensor
from torch.nn.init import xavier_uniform_

from ..util.constants import DEFAULT_FEATURES
from ..util.util import mlp


class _PreLNTransformerEncoder(nn.TransformerEncoderLayer):
    def forward(self, src: Tensor, src_mask: Optional[Tensor] = None,
                src_key_padding_mask: Optional[Tensor] = None) -> Tensor:
        """Simple override of TransformerEncoderLayer to do Pre-LN (https://arxiv.org/abs/2002.04745)"""
        src2 = self.norm1(src)
        src2 = self.self_attn(src2, src2, src2, attn_mask=src_mask,
                              key_padding_mask=src_key_padding_mask)[0]
        src = src + self.dropout1(src2)
        src2 = self.norm2(src)
        src2 = self.linear2(self.dropout(self.activation(self.linear1(src2))))
        src = src + self.dropout2(src2)
        return src


class EARL(nn.Module):
    def __init__(
            self,
            n_dims: int = 128,
            n_layers: int = 4,
            n_heads: int = 4,
            n_preprocess_layers: int = 1,
            n_postprocess_layers: int = 0,
            dropout: float = 0.,
            n_features: Optional[int] = None,
            dim_feedforward: Optional[int] = None,
            pre_ln=True
    ):
        """
        Create an EARL (Extensible Attention-based Rocket League) model

        :param n_dims: number of dimensions in the intermediate and output representations.
        :param n_layers: number of encoder layers.
        :param n_heads: number of heads in encoder layers.
        :param n_preprocess_layers: number of dense layers before encoder.
        :param n_postprocess_layers: number of dense layers after encoder.
        :param dropout: dropout for encoder layers.
        :param n_features: number of features in the input (last dimension).
        :param pre_ln: whether or not to do pre-layer norm instead of post (see https://arxiv.org/abs/2002.04745)
        """
        super().__init__()
        self.n_features = n_features or len(DEFAULT_FEATURES)
        self.n_dims = n_dims
        self.n_layers = n_layers
        self.n_heads = n_heads
        self.pre_ln = pre_ln

        self.preprocess = mlp(self.n_features, n_dims, n_preprocess_layers)
        if n_postprocess_layers > 0:
            self.postprocess = mlp(n_dims, n_dims, n_postprocess_layers - 1, n_dims)
        else:
            self.postprocess = nn.Identity()

        if dim_feedforward is None:
            dim_feedforward = 2 * n_dims

        encoder_layer = _PreLNTransformerEncoder if pre_ln else nn.TransformerEncoderLayer
        self.blocks = nn.ModuleList([
            encoder_layer(n_dims, n_heads, dim_feedforward, dropout=dropout, batch_first=True)
            for _ in range(n_layers)
        ])
        self._reset_parameters()

    def _reset_parameters(self):
        r"""Initiate parameters in the transformer model. Taken from PyTorch Transformer impl"""
        for p in self.blocks.parameters():
            if p.dim() > 1:
                xavier_uniform_(p)

    def forward(self, entities: torch.Tensor, mask: Optional[torch.Tensor] = None):
        emb = self.initial_dense(entities)
        emb = self.preprocess(emb)
        for block in self.blocks:
            emb = block(emb, src_key_padding_mask=mask)
        emb = self.postprocess(emb)
        return emb

    def __repr__(self):
        return f"EARL(n_dims={self.n_dims},n_layers={self.n_layers},n_heads={self.n_heads})"



from rlgym.utils import TerminalCondition
from rlgym.utils.gamestates import GameState


class GameCondition(TerminalCondition):  # Mimics a Rocket League game
    def __init__(self, tick_skip=8, seconds_left=300, seconds_per_goal_forfeit=None,
                 max_overtime_seconds=float("inf"), max_no_touch_seconds=float("inf")):
        # NOTE: Since game isn't reset to kickoff by default,
        # you need to keep this outside the main loop as well,
        # checking the done variable after each terminal
        super().__init__()
        self.tick_skip = tick_skip
        self.seconds_left = seconds_left
        self.timer = seconds_left
        self.overtime = False
        self.done = True
        self.initial_state = None
        self.seconds_per_goal_forfeit = seconds_per_goal_forfeit  # SPG = Seconds Per Goal
        self.max_overtime = max_overtime_seconds
        self.max_no_touch = max_no_touch_seconds
        self.last_touch = None
        self.differential = None

    def reset(self, initial_state: GameState):
        if self.done:  # New game
            self.timer = self.seconds_left
            self.initial_state = initial_state
            self.overtime = False
            self.done = False
            self.last_touch = self.seconds_left
            self.differential = 0

    def is_terminal(self, current_state: GameState) -> bool:
        reset = False

        differential = ((current_state.blue_score - self.initial_state.blue_score)
                        - (current_state.orange_score - self.initial_state.orange_score))

        if differential != self.differential:  # Goal scored
            reset = True

        if self.overtime:
            if differential != 0:
                self.done = True
            elif self.timer >= self.max_overtime:
                self.done = True  # Call it a draw
            else:
                self.timer += self.tick_skip / 120
                self.done = False
        else:
            if self.timer <= 0 and current_state.ball.position[2] <= 110:
                # Can't detect ball on ground directly, should be an alright approximation.
                # Anything below z vel of ~690uu/s should be detected. 50% for 1380 etc.
                if differential != 0:
                    self.done = True
                else:
                    self.overtime = True
                    self.last_touch = self.timer  # Just for convenience
                    self.done = False
                    reset = True
            elif (self.seconds_per_goal_forfeit is not None
                  and abs(differential) >= 3
                  and self.timer / abs(differential) < self.seconds_per_goal_forfeit):
                # Too few seconds per goal to realistically win
                self.done = True
            else:
                self.timer -= self.tick_skip / 120

        if abs(self.last_touch - self.timer) >= self.max_no_touch:
            self.done = True
        elif any(p.ball_touched for p in current_state.players):
            self.last_touch = self.timer

        self.differential = differential

        return reset or self.done



from all.environments import MultiagentEnvironment, MultiagentPettingZooEnv

from rlgym.gym import Gym
from rlgym_tools.pettingzoo_utils import PettingZooEnv


def autonomous_env(env: Gym):
    return MultiagentPettingZooEnv(PettingZooEnv(env), name="rlgym_pettingzoo")


# class AutonomousEnv(MultiagentEnvironment):
#     def __init__(self, env: Gym):
#         self.env = env
#         self.
#
#     def reset(self):
#         return self.env.reset()
#
#     def step(self, action):
#         pass
#
#     def render(self, **kwargs):
#         pass
#
#     def close(self):
#         pass
#
#     def agent_iter(self):
#         pass
#
#     def last(self):
#         pass
#
#     def is_done(self, agent):
#         pass
#
#     @property
#     def name(self):
#         pass
#
#     @property
#     def state_spaces(self):
#         pass
#
#     @property
#     def action_spaces(self):
#         pass
#
#     @property
#     def device(self):
#         pass



import numpy as np
from pettingzoo import AECEnv

from rlgym.gym import Gym


class PettingZooEnv(AECEnv):
    """
    Wrapper for using the RLGym env with PettingZoo,
    """

    def __init__(self, env: Gym):
        """
        :param env: the environment to wrap.
        """
        super().__init__()
        self.env = env
        self.metadata = env.metadata
        self.agents = list(range(self.env._match.agents))
        self.possible_agents = self.agents.copy()

        self.observation_spaces = {agent: self.env.observation_space for agent in self.agents}
        self.action_spaces = {agent: self.env.action_space for agent in self.agents}

        self._reset_values()

    def _reset_values(self):
        self.rewards = {agent: 0 for agent in self.agents}
        self._cumulative_rewards = {agent: 0 for agent in self.agents}
        self.dones = {agent: False for agent in self.agents}
        self.infos = {agent: {} for agent in self.agents}
        self.state = {agent: None for agent in self.agents}
        self.observations = {agent: None for agent in self.agents}
        self.actions = {}  # For storing until we have enough actions to do an in-game step

        # Somewhat redundant, but would support any type of agent
        self._current_agent_index = 0
        self.agent_selection = self.agents[0]

    def reset(self):
        self._reset_values()

        observations = self.env.reset()
        self.observations = dict(zip(self.agents, observations))
        return self.observations

    def step(self, action):
        agent = self.agent_selection

        self.actions[agent] = action

        if self.agent_selection == self.agents[-1]:  # Only apply once everyone has registered
            action_array = np.stack([self.actions[agent] for agent in self.agents])
            observations, rewards, done, info = self.env.step(action_array)
            assert len(observations) == len(rewards) == self.num_agents
            self.observations = dict(zip(self.agents, observations))
            self.rewards = dict(zip(self.agents, rewards))
            self.dones = {agent: done for agent in self.agents}
            self.infos = {agent: info for agent in self.agents}
        self._current_agent_index = (self.agent_selection + 1) % self.num_agents
        self.agent_selection = self.agents[self._current_agent_index]

    def observe(self, agent):
        return self.observations[agent]

    def render(self, mode='human'):
        self.env.render(mode)

    def state(self):
        raise NotImplementedError

    def seed(self, seed=None):
        self.env.seed(seed)

    def close(self):
        self.env.close()


def parallel_pettingzoo_env(env: Gym):
    # Preliminary solution for making a ParallelEnv
    import supersuit as ss
    return ss.to_parallel(PettingZooEnv(env))



from typing import Any

import gym
import numpy as np
from gym.spaces import Discrete
from rlgym.utils.action_parsers import ActionParser
from rlgym.utils.gamestates import GameState


class LookupAction(ActionParser):
    def __init__(self, bins=None):
        super().__init__()
        if bins is None:
            self.bins = [(-1, 0, 1)] * 5
        elif isinstance(bins, int):
            assert bins % 2 == 1, "Need an odd number of bins"
            self.bins = [tuple(np.linspace(-1, 1, bins))] * 5
        elif isinstance(bins[0], (float, int)):
            self.bins = [bins] * 5
        else:
            assert len(bins) == 5, "Need bins for throttle, steer, pitch, yaw and roll"
            self.bins = bins
        self.lookup_table = self.make_lookup_table(self.bins)

    @staticmethod
    def make_lookup_table(bins):
        actions = []
        # Ground
        for throttle in bins[0]:
            for steer in bins[1]:
                for boost in (0, 1):
                    for handbrake in (0, 1):
                        if boost == 1 and throttle != 1:
                            continue
                        actions.append([throttle or boost, steer, 0, steer, 0, 0, boost, handbrake])
        # Aerial
        for pitch in bins[2]:
            for yaw in bins[3]:
                for roll in bins[4]:
                    for jump in (0, 1):
                        for boost in (0, 1):
                            if jump == 1 and yaw != 0:
                                continue  # Only need roll for sideflip
                            if pitch == roll == jump == 0:
                                continue  # Duplicate with ground

                            is_flip = jump == 1 and not pitch == roll == 0
                            if is_flip and abs(pitch) != 1 and abs(roll) != 1:
                                continue  # Deadzone is inconsistent, only allow flips if pitch/yaw is fully pressed

                            handbrake = is_flip  # Enable handbrake for potential wavedashes
                            actions.append([boost, yaw, pitch, yaw, roll, jump, boost, handbrake])
        actions = np.array(actions)
        return actions

    def get_action_space(self) -> gym.spaces.Space:
        return Discrete(len(self.lookup_table))

    def parse_actions(self, actions: Any, state: GameState) -> np.ndarray:
        indexes = np.array(actions, dtype=np.int32)
        indexes = np.squeeze(indexes)
        return self.lookup_table[indexes]



import multiprocessing as mp
import os
import time
from typing import Optional, List, Union, Any, Callable, Sequence

import numpy as np
from stable_baselines3.common.vec_env import SubprocVecEnv, CloudpickleWrapper, VecEnv
from stable_baselines3.common.vec_env.base_vec_env import (
    VecEnvObs,
    VecEnvStepReturn,
    VecEnvIndices,
)
from stable_baselines3.common.vec_env.subproc_vec_env import _worker

from rlgym.envs import Match
from rlgym.gym import Gym
from rlgym.gamelaunch import LaunchPreference

class SB3MultipleInstanceEnv(SubprocVecEnv):
    """
    Class for launching several Rocket League instances into a single SubprocVecEnv for use with Stable Baselines.
    """

    MEM_INSTANCE_LAUNCH = 3.5e9
    MEM_INSTANCE_LIM = 4e6

    @staticmethod
    def estimate_supported_processes():
        import psutil

        vm = psutil.virtual_memory()
        # Need 3.5GB to launch, reduces to 350MB after a while
        est_proc_mem = round(
            (vm.available - SB3MultipleInstanceEnv.MEM_INSTANCE_LAUNCH)
            / SB3MultipleInstanceEnv.MEM_INSTANCE_LAUNCH
        )
        est_proc_cpu = os.cpu_count()
        est_proc = min(est_proc_mem, est_proc_cpu)
        return est_proc

    def __init__(
        self,
        match_func_or_matches: Union[Callable[[], Match], Sequence[Match]],
        num_instances: Optional[int] = None,
        launch_preference: Optional[Union[LaunchPreference, str]] = LaunchPreference.EPIC,
        wait_time: float = 30,
        force_paging: bool = False,
        auto_minimize: bool = True
    ):
        """
        :param match_func_or_matches: either a function which produces the a Match object, or a list of Match objects.
                                Needs to be a function so that each subprocess can call it and get their own objects.
        :param num_instances: the number of Rocket League instances to start up,
                              or "auto" to estimate how many instances are supported (requires psutil).
        :param wait_time: the time to wait between launching each instance. Default one minute.
        :param force_paging: enable forced paging of each spawned rocket league instance to reduce memory utilization
                             immediately, instead of allowing the OS to slowly page untouched allocations.
                             WARNING: This will require you to potentially expand your Windows Page File, and it may
                             substantially increase disk activity, leading to decreased disk lifetime.
                             Use at your own peril.
                             https://www.tomshardware.com/news/how-to-manage-virtual-memory-pagefile-windows-10,36929.html
                             Default is off: OS dictates the behavior.
        :param auto_minimize: Automatically minimize the game window when launching Rocket League
        """
        if callable(match_func_or_matches):
            assert num_instances is not None, (
                "If using a function to generate Match objects, "
                "num_instances must be specified"
            )
            if num_instances == "auto":
                num_instances = SB3MultipleInstanceEnv.estimate_supported_processes()
            match_func_or_matches = [
                match_func_or_matches() for _ in range(num_instances)
            ]

        def get_process_func(i):
            def spawn_process():
                match = match_func_or_matches[i]
                env = Gym(
                    match,
                    pipe_id=os.getpid(),
                    launch_preference=launch_preference,
                    use_injector=True,
                    force_paging=force_paging,
                    auto_minimize=auto_minimize
                )
                return env

            return spawn_process

        # super().__init__([])  Super init intentionally left out since we need to launch processes with delay

        env_fns = [get_process_func(i) for i in range(len(match_func_or_matches))]

        # START - Code from SubprocVecEnv class
        self.waiting = False
        self.closed = False
        n_envs = len(env_fns)

        # Fork is not a thread safe method (see issue #217)
        # but is more user friendly (does not require to wrap the code in
        # a `if __name__ == "__main__":`)
        forkserver_available = "forkserver" in mp.get_all_start_methods()
        start_method = "forkserver" if forkserver_available else "spawn"
        ctx = mp.get_context(start_method)

        self.remotes, self.work_remotes = zip(*[ctx.Pipe() for _ in range(n_envs)])
        self.processes = []
        for work_remote, remote, env_fn in zip(
            self.work_remotes, self.remotes, env_fns
        ):
            args = (work_remote, remote, CloudpickleWrapper(env_fn))
            # daemon=True: if the main process crashes, we should not cause things to hang
            process = ctx.Process(
                target=_worker, args=args, daemon=True
            )  # pytype:disable=attribute-error
            process.start()
            self.processes.append(process)
            work_remote.close()

            if len(self.processes) != len(env_fns):
                time.sleep(wait_time)  # ADDED - Waits between starting Rocket League instances

        self.remotes[0].send(("get_spaces", None))
        observation_space, action_space = self.remotes[0].recv()
        # END - Code from SubprocVecEnv class

        self.n_agents_per_env = [m.agents for m in match_func_or_matches]
        self.num_envs = sum(self.n_agents_per_env)
        VecEnv.__init__(self, self.num_envs, observation_space, action_space)

    def reset(self) -> VecEnvObs:
        for remote in self.remotes:
            remote.send(("reset", None))

        flat_obs = []
        for remote, n_agents in zip(self.remotes, self.n_agents_per_env):
            obs = remote.recv()
            if n_agents <= 1:
                flat_obs.append(obs)
            else:
                flat_obs += obs
        return np.asarray(flat_obs)

    def step_async(self, actions: np.ndarray) -> None:
        i = 0
        for remote, n_agents in zip(self.remotes, self.n_agents_per_env):
            remote.send(("step", actions[i : i + n_agents]))
            i += n_agents
        self.waiting = True

    def step_wait(self) -> VecEnvStepReturn:
        flat_obs = []
        flat_rews = []
        flat_dones = []
        flat_infos = []
        for remote, n_agents in zip(self.remotes, self.n_agents_per_env):
            obs, rew, done, info = remote.recv()
            if n_agents <= 1:
                flat_obs.append(obs)
                flat_rews.append(rew)
                flat_dones.append(done)
                flat_infos.append(info)
            else:
                flat_obs += obs
                flat_rews += rew
                flat_dones += [done] * n_agents
                flat_infos += [info] * n_agents
        self.waiting = False
        return (
            np.asarray(flat_obs),
            np.array(flat_rews),
            np.array(flat_dones),
            flat_infos,
        )

    def seed(self, seed: Optional[int] = None) -> List[Union[None, int]]:
        res = super(SB3MultipleInstanceEnv, self).seed(seed)
        return sum([r] * a for r, a in zip(res, self.n_agents_per_env))

    def _get_target_remotes(self, indices: VecEnvIndices) -> List[Any]:
        # Override to prevent out of bounds
        indices = self._get_indices(indices)
        remotes = []
        for i in indices:
            tot = 0
            for remote, n_agents in zip(self.remotes, self.n_agents_per_env):
                tot += n_agents
                if i < tot:
                    remotes.append(remote)
                    break
        return remotes



from typing import Tuple
from ray.rllib import MultiAgentEnv
from ray.rllib.utils import override
from ray.rllib.utils.typing import MultiAgentDict
from rlgym.gym import Gym

# @ray.remote(num_cpus=1, memory=3.5e9)
class RLLibEnv(MultiAgentEnv):
    def __init__(self, env: Gym):
        super().__init__()
        self.env = env
        self.observation_space = env.observation_space
        self.action_space = env.action_space
        self._agent_ids = [i for i in range(env._match.agents)]

    @override(MultiAgentEnv)
    def reset(self) -> MultiAgentDict:
        observations = self.env.reset()
        return {i: obs for i, obs in enumerate(observations)}

    @override(MultiAgentEnv)
    def step(self, action_dict: MultiAgentDict) -> \
            Tuple[MultiAgentDict, MultiAgentDict, MultiAgentDict, MultiAgentDict]:
        action_array = [val for val in action_dict.values()]
        observations, rewards, done, info = self.env.step(action_array)

        obs_dict = {}
        rew_dict = {}
        done_dict = {"__all__": done}
        info_dict = {}
        for i, (obs, rew) in enumerate(zip(observations, rewards)):
            obs_dict[i] = obs
            rew_dict[i] = rew
            done_dict[i] = done
            info_dict[i] = info
        return obs_dict, rew_dict, done_dict, info_dict

    @override(MultiAgentEnv)
    def action_space_sample(self, agent_ids: list = None) -> MultiAgentDict:
        return {agent_id:self.action_space.sample() for agent_id in self._agent_ids}



import numpy as np

from rlgym.utils import RewardFunction
from rlgym.utils.common_values import BLUE_TEAM, BACK_WALL_Y
from rlgym.utils.gamestates import PlayerData, GameState


class KickoffReward(RewardFunction):
    def reset(self, initial_state: GameState):
        pass

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        if player.team_num == BLUE_TEAM:
            player_data = player.car_data
            ball_data = state.ball
        else:
            player_data = player.inverted_car_data
            ball_data = state.inverted_ball
        return ball_data.position[1] / BACK_WALL_Y


class KickoffTerminal(RewardFunction):
    def reset(self, initial_state: GameState):
        pass

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        pass


kickoff_config = dict(
    reward_function=ConstantReward(),
    terminal_conditions=[AirDribbleTerminal()],
    state_setter=AirDribbleState(),
    team_size=1,
    spawn_opponents=True
)



from rlgym.utils.reward_functions import RewardFunction
from rlgym.utils.reward_functions.common_rewards.player_ball_rewards import VelocityPlayerToBallReward

from rlgym.utils.gamestates import GameState, PlayerData

from numpy import ndarray


class KickoffReward(RewardFunction):
    """
    a simple reward that encourages driving towards the ball fast while it's in the neutral kickoff position
    """
    def __init__(self):
        super().__init__()
        self.vel_dir_reward = VelocityPlayerToBallReward()

    def reset(self, initial_state: GameState):
        self.vel_dir_reward.reset(initial_state)

    def get_reward(
        self, player: PlayerData, state: GameState, previous_action: ndarray
    ) -> float:
        reward = 0
        if state.ball.position[0] == 0 and state.ball.position[1] == 0:
            reward += self.vel_dir_reward.get_reward(player, state, previous_action)
        return reward



import setuptools
from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='EARL-pytorch',
    version='0.5.2',
    packages=setuptools.find_packages(),
    install_requires=[
        "torch>=1.9.0",
        "numpy>=1.19",
    ],
    url='https://github.com/Rolv-Arild/EARL-pytorch',
    license='MIT License',
    author='Rolv-Arild Braaten',
    author_email='rolv_arild@hotmail.com',
    description='EARL - Extensible Attention-based Rocket League model',
    # long_description=long_description,
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



from rlgym.utils.state_setters import StateSetter
from rlgym.utils.state_setters import StateWrapper
from rlgym.utils.common_values import BALL_RADIUS, CEILING_Z, BLUE_TEAM
import numpy as np
from numpy import random as rand

X_MAX = 7000
Y_MAX = 9000
Z_MAX_CAR = 1900
PITCH_MAX = np.pi / 2
ROLL_MAX = np.pi


class KickoffLikeSetter(StateSetter):

    def __init__(self, cars_on_ground: bool = True, ball_on_ground: bool = True):
        """
        RandomState constructor.


        :param cars_on_ground: Boolean indicating whether cars should only be placed on the ground.
        :param ball_on_ground: Boolean indicating whether ball should only be placed on the ground.
        """
        super().__init__()
        self.cars_on_ground = cars_on_ground
        self.ball_on_ground = ball_on_ground
        self.yaw_vector = np.asarray([-1, 0, 0])

    def reset(self, state_wrapper: StateWrapper):
        """
        Modifies the StateWrapper to contain random values the ball and each car.

        :param state_wrapper: StateWrapper object to be modified with desired state values.
        """
        self._reset_ball_random(state_wrapper, self.ball_on_ground)
        self._reset_cars_random(state_wrapper, self.cars_on_ground)

    def _reset_ball_random(self, state_wrapper: StateWrapper, ball_grounded):
        """
        Function to set the ball to a random position.

        :param state_wrapper: StateWrapper object to be modified.
        :param ball_grounded: Boolean indicating whether ball should only be placed on the ground.
        """
        state_wrapper.ball.set_pos(rand.random() * X_MAX - X_MAX / 2, 0,
                                   BALL_RADIUS if ball_grounded else rand.random() * (
                                       CEILING_Z - 2 * BALL_RADIUS) + BALL_RADIUS)

    def _reset_cars_random(self, state_wrapper: StateWrapper, on_ground: bool):
        """
        Function to set all cars to a random position.

        :param state_wrapper: StateWrapper object to be modified.
        :param on_ground: Boolean indicating whether to place cars only on the ground.
        """
        for i, car in enumerate(state_wrapper.cars):
            # set random position and rotation for all cars based on pre-determined ranges
            if car.team_num == BLUE_TEAM:

                car.set_pos(rand.random() * X_MAX - X_MAX / 2, -abs(rand.random() * Y_MAX - Y_MAX / 2),
                            rand.random() * Z_MAX_CAR + 150)

                # compute vector from ball to car
                rel_ball_car_vector = car.position - state_wrapper.ball.position

                # calculate the angle between the yaw vector and relative vector and use that as yaw.
                yaw = np.arccos(np.dot(rel_ball_car_vector / np.linalg.norm(rel_ball_car_vector), self.yaw_vector))

                # then sprinkle in more variation by offsetting by random angle up to pi/8 radians (this is arbitrary)
                max_offset_angle = np.pi / 8
                yaw += rand.random() * max_offset_angle * 2 - max_offset_angle

                car.set_rot(rand.random() * PITCH_MAX - PITCH_MAX / 2, yaw, rand.random() * ROLL_MAX - ROLL_MAX / 2)

                car.boost = np.random.uniform(0.2, 1)

                # 100% of cars will be set on ground if on_ground == True
                # otherwise, 50% of cars will be set on ground
                if on_ground or rand.random() < 0.5:
                    # z position (up/down) is set to ground
                    car.set_pos(z=17)

                    # pitch (front of car up/down) set to 0
                    # roll (side of car up/down) set to 0
                    car.set_rot(pitch=0, roll=0, yaw=yaw)
                # x angular velocity (affects pitch) set to 0
                # y angular velocity (affects roll) set to 0
                # z angular velocity (affects yaw) set to 0
                car.set_ang_vel(x=0, y=0, z=0)
                car.set_lin_vel(x=0, y=0, z=0)
            else:
                # the cars in state_wrapper.cars are in order, starting with blue so we can compute blue first and then
                # just copy to orange cars, to assure symmetry
                car_to_copy = state_wrapper.cars[i - len(state_wrapper.cars) // 2]
                car.set_pos(car_to_copy.position[0], -car_to_copy.position[1], car_to_copy.position[2])
                car.set_rot(car_to_copy.rotation[0], -car_to_copy.rotation[1], -car_to_copy.rotation[2])
                car.boost = car_to_copy.boost
                car.set_ang_vel(x=0, y=0, z=0)
                car.set_lin_vel(x=0, y=0, z=0)



import copy
import math
import os
from typing import Any, Union, Tuple, List, Dict

import gym.spaces
import numpy as np
from rlgym.envs import Match
from rlgym.gym import Gym
from rlgym.utils import RewardFunction, ObsBuilder, StateSetter, TerminalCondition
from rlgym.utils.action_parsers import ActionParser, DefaultAction
from rlgym.utils.common_values import BLUE_TEAM
from rlgym.utils.gamestates import PlayerData, GameState
from rlgym.utils.obs_builders import DefaultObs
from rlgym.utils.reward_functions import DefaultReward
from rlgym.utils.state_setters import StateWrapper, DefaultState
from rlgym.utils.terminal_conditions.common_conditions import TimeoutCondition, GoalScoredCondition


def get_reduced_state(b, o, state):
    reduced_state = copy.deepcopy(state)
    reduced_state.players = (reduced_state.players[:b]  # Blue
                             + reduced_state.players[3:3 + o])  # Orange
    return reduced_state


class FilteredRewTermObsAct(RewardFunction, TerminalCondition, ObsBuilder, ActionParser, StateSetter):
    def __init__(
            self,
            team_sizes: list,
            reward_fn: RewardFunction,
            terminal_conditions: List[TerminalCondition],
            obs_builder: ObsBuilder,
            action_parser: ActionParser,
            state_setter: StateSetter
    ):
        super().__init__()
        self.team_sizes = team_sizes
        self.reward_fn = reward_fn
        self.terminal_conditions = terminal_conditions
        self.obs_builder = obs_builder
        self.action_parser = action_parser
        self.state_setter = state_setter

        self.current_state = None
        self.reduced_state = None

    def _update_state(self, state: GameState):
        if state != self.current_state:
            self.current_state = state
            self.reduced_state = get_reduced_state(*self.team_sizes, state)

    def reset(self, initial_state_or_wrapper: Union[GameState, StateWrapper]):
        # Reset in StateSetter has a different signature to the others
        if isinstance(initial_state_or_wrapper, StateWrapper):
            state_wrapper = initial_state_or_wrapper
            b, o = self.team_sizes
            reduced_wrapper = StateWrapper(3, 3)

            # Now pass original objects so values change in original wrapper
            reduced_wrapper.ball = state_wrapper.ball
            reduced_wrapper.cars = state_wrapper.cars[:b] + state_wrapper.cars[3:3 + o]
            self.state_setter.reset(reduced_wrapper)

            bc = oc = 0
            for car in state_wrapper.cars[b:3] + state_wrapper.cars[3 + o:]:
                if car.team_num == BLUE_TEAM:
                    bc += 1
                    car.set_pos((-1) ** bc * (bc // 2) * 100, -6200, 100)
                    car.set_rot(0, math.pi / 2, 0)
                else:
                    oc += 1
                    car.set_pos((-1) ** oc * (oc // 2) * 100, 6200, 100)
                    car.set_rot(0, -math.pi / 2, 0)
                car.set_lin_vel(0, 0, 0)
                car.set_ang_vel(0, 0, 0)
        elif initial_state_or_wrapper != self.current_state:
            self._update_state(initial_state_or_wrapper)
            assert self.reduced_state is not None
            self.reward_fn.reset(self.reduced_state)
            for tc in self.terminal_conditions:
                tc.reset(self.reduced_state)
            self.obs_builder.reset(self.reduced_state)

    def get_reward(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> float:
        self._update_state(state)
        if player.car_id not in (p.car_id for p in self.reduced_state.players):
            return float("nan")
        return self.reward_fn.get_reward(player, self.reduced_state, previous_action)

    def build_obs(self, player: PlayerData, state: GameState, previous_action: np.ndarray) -> Any:
        self._update_state(state)
        if player.car_id not in (p.car_id for p in self.reduced_state.players):
            return None
        return self.obs_builder.build_obs(player, self.reduced_state, previous_action)

    def is_terminal(self, current_state: GameState) -> bool:
        self._update_state(current_state)
        return any(tc.is_terminal(self.reduced_state) for tc in self.terminal_conditions)

    def get_action_space(self) -> gym.spaces.Space:
        return self.action_parser.get_action_space()

    def parse_actions(self, actions: Any, state: GameState) -> np.ndarray:
        assert len(actions) == sum(self.team_sizes)
        actions = self.action_parser.parse_actions(actions, state)
        new_actions = np.zeros((6, 8))
        b = o = 0
        for act in actions:
            if b < self.team_sizes[0]:
                new_actions[b, :] = act
                b += 1
            else:
                new_actions[3 + o, :] = act
                o += 1
        return new_actions


class FilteredGym(Gym):
    """
    The FilteredGym creates a 3v3 instance and sets gamemode by spawning them outside the field.
    This adds the ability to change gamemode within the same instance,
    as well as unfair matchups which are normally impossible (1v2, 1v3, 2v3, 0v2, 0v3)
    """

    def __init__(self, blue_players: int, orange_players: int, **match_kwargs):
        self._team_sizes = [blue_players, orange_players]
        self._filter = FilteredRewTermObsAct(
            self._team_sizes,
            match_kwargs.pop("reward_function", DefaultReward()),
            match_kwargs.pop("terminal_conditions", [TimeoutCondition(225), GoalScoredCondition()]),
            match_kwargs.pop("obs_builder", DefaultObs()),
            match_kwargs.pop("action_parser", DefaultAction()),
            match_kwargs.pop("state_setter", DefaultState())
        )
        match = Match(
            reward_function=self._filter,
            terminal_conditions=self._filter,
            obs_builder=self._filter,
            action_parser=self._filter,
            state_setter=self._filter,
            team_size=3,
            self_play=True,
            **match_kwargs
        )
        super().__init__(match, os.getpid(), use_injector=True)

    def reset(self, return_info=False, blue_players=None, orange_players=None) -> Union[List, Tuple]:
        if blue_players is not None:
            self._team_sizes[0] = blue_players
        if orange_players is not None:
            self._team_sizes[1] = orange_players
        if return_info:
            obs, info = super(FilteredGym, self).reset(True)
            state = info.get("state")
            if state is not None:
                info["state"] = get_reduced_state(*self._team_sizes, state)
            return obs, info
        return super(FilteredGym, self).reset(False)

    def step(self, actions: Any) -> Tuple[List, List, bool, Dict]:
        obs, reward, done, info = super(FilteredGym, self).step(actions)
        b, o = self._team_sizes
        state = info.get("state")
        if state is not None:
            info["state"] = get_reduced_state(b, o, state)
        return obs[:b] + obs[3:3 + o], reward[:b] + reward[3:3 + o], done, info


if __name__ == '__main__':
    env = FilteredGym(1, 2, game_speed=1)

    while True:
        done = False
        env.reset()
        while not done:
            act = [env.action_space.sample() for _ in range(3)]
            obs, rew, done, info = env.step(act)



# From https://github.com/RLBot/RLBot/blob/master/src/main/python/rlbot/version.py
# Store the version here so:
# 1) we don't load dependencies by storing it in __init__.py
# 2) we can import it in setup.py for the same reason
# 3) we can import it into your module module
# https://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package

__version__ = '1.8.2'

release_notes = {
    '1.8.2': """
       - Fix no touch timer in GameCondition (Rolv)
       - Update RLLib example (Aech)
    """,
    '1.8.1': """
        - Refactor GameCondition (Rolv, Impossibum)
        - Fix a small mistake in LookupAction (Rolv)
    """,
    '1.8.0': """
        - Add lookup parser as used by Nexto/Tecko (Rolv)
        - Add customizable odds to WallPracticeState (Soren)
        - Add code for reducing SB3 model size for RLBot botpack, with example (DI3D)
        - Update AdvancedPadder for RLGym 1.2 (Kaiyotech)
        - Update example code for RLGym 1.2 (mstuettgen)
        - Fix AdvancedStacker (Some Rando)
        - Fix broken imports in SequentialRewards (Some Rando)
        - Fix bad indent in JumpTouchReward (Some Rando)
        """,
    '1.7.0':
        """
        - Add AdvancedObsPadder (Impossibum)
        - Add JumpTouchReward (Impossibum)
        - Fix NameError in KickoffReward (benjamincburns)
        - Add generate_probabilities as a method to ReplaySetter (Rolv)
        - Upgrade WallSetter (Soren) and fix bug when num_cars == 1 (Kaiyotech)
        - Add max overtime to GameCondition (Rolv)
        """,
    '1.6.6':
        """
        -WallStateSetter now has airdribble setups and harder wall plays
        """,
    '1.6.5':
        """
        -GoalieStateSetter now better emulates incoming airdribbles
        -fixed WallStateSetter bug and increased starting diversity
        """,
    '1.6.4':
        """
        -Added wall play state setter (Soren)
        """,
    '1.6.3':
        """
        - Fix hoops-like setter for multiple players (Carrot)
        """,
    '1.6.2':
        """
        - Added hoops-like setter (Carrot)
        - Fixed kickoff-like setter (Carrot)
        - Fixed pitch in KBMAction (Rolv)
        - Added include_frame option in replay converter (wrongu)
        """,
    '1.6.1':
        """
        - Fixed angular velocities in replay augmenter
        """,
    '1.6.0':
        """
        - Added GoaliePracticeState setter (Soren)
        - Added ReplaySetter (Carrot)
        - Added AugmentSetter (NeverCast)
        - Fixed an error in replay converter
        """,
    '1.5.3':
        """
        - Yet another fix for GameCondition
        """,
    '1.5.2':
        """
        - Another fix for GameCondition
        """,
    '1.5.1':
        """
        - Fix for GameCondition
        """,
    '1.5.0':
        """
        - Add SB3CombinedLogReward (LuminousLamp)
        - Add SB3InstantaneousFPSCallback (LuminousLamp)
        - Add KickoffLikeSetter (Carrot)
        - Add GameCondition (Rolv)
        """,
    '1.4.1':
        """
        - Remove SB3MultiDiscreteWrapper
        - Update SB3 examples to include action parsers
        """,
    '1.4.0':
        """
        - Add KBM action parser
        """,
    '1.3.0':
        """
        - Add KickoffReward (Impossibum)
        - Add SequentialReward (Impossibum)
        - Better individual reward logging for SB3 (LuminousLamp)
        - Add launch_preference as parameter to SB3MultipleInstanceEnv
        """,
    '1.2.0':
        """
        - Added multi-model support for SB3
        - Added weighted sample setter
        - Added general stacker
        - Various bug fixes
        """,
    '1.1.4':
        """
        - Bug fixes for AdvancedStacker
        """,
    '1.1.3':
        """
        - Fix RLGym 1.0 incompatibility
        """,
    '1.1.2':
        """
        - Fix version import (was absolute, now relative)
        """,
    '1.1.1':
        """
        - Fixed an invalid Python 3.7 import
        - AnnealRewards now has initial_count parameter
        """,
    '1.1.0':
        """
        - Added functionality in SB3MultipleInstanceEnv to take multiple match objects (one for each instance).
        - Added AnnealRewards for transitioning between reward functions.
        - Added granularity option to SB3MultiDiscreteWrapper.
        - Added negative slope parameter to DiffReward (negative values are multiplied by this).
        """,
    '1.0.0':
        """
        - Added replay to rlgym GameState converter
        - Moved SB3 environments from rlgym (now called SB3SingleInstanceEnv and SB3MultipleInstanceEnv) and fixed some bugs
        - Added SB3MultiDiscreteWrapper, SB3DistributeRewardsWrapper and SB3LogReward 
        - Added extra reward functions (DiffReward, DistributeRewards and MultiplyRewards)
        - Added RLLibEnv
        - Added working example code for SB3 and RLlib
        """
}


def get_current_release_notes():
    if __version__ in release_notes:
        return release_notes[__version__]
    return ''


def print_current_release_notes():
    print(f"Version {__version__}")
    print(get_current_release_notes())
    print("")



import math
from copy import deepcopy
from random import getrandbits, shuffle
from typing import List

import numpy as np
from rlgym.utils.state_setters.state_setter import StateSetter
from rlgym.utils.state_setters.wrappers import CarWrapper
from rlgym.utils.state_setters.wrappers import StateWrapper

PI = math.pi


class AugmentSetter(StateSetter):
    MASK_SWAP_FRONT_BACK = 0b01
    MASK_SWAP_LEFT_RIGHT = 0b10

    def __init__(self, state_setter: StateSetter, shuffle_within_teams=True, swap_front_back=True,
                 swap_left_right=True):
        super().__init__()
        self.state_setter = state_setter
        self.shuffle_within_teams = shuffle_within_teams
        self.swap_front_back = swap_front_back
        self.swap_left_right = swap_left_right

    def reset(self, state_wrapper: StateWrapper):
        self.state_setter.reset(state_wrapper)
        # self._debug(state_wrapper)

        bits = getrandbits(2)
        if self.shuffle_within_teams:
            self.shuffle_players(state_wrapper)

        if self.swap_front_back and (bits & AugmentSetter.MASK_SWAP_FRONT_BACK):
            self.mirror_front_back(state_wrapper)

        if self.swap_left_right and (bits & AugmentSetter.MASK_SWAP_LEFT_RIGHT):
            self.mirror_left_right(state_wrapper)

        # self._debug(state_wrapper)

    def _debug(self, state_wrapper: StateWrapper):
        print("\n".join(f"Car {car.id}, team: {car.team_num}, pos: {car.position}" for car in state_wrapper.cars))
        ball = state_wrapper.ball
        print(f"Ball pos: {ball.position}")

    @staticmethod
    def _map_cars(cars0: List[CarWrapper], cars1: List[CarWrapper]):
        # Transfer all the states from cars0 to cars1
        for car0, car1 in zip(cars0, cars1):
            if cars0 == car1:  # By reference
                continue
            car0.position[:], car1.position[:] = \
                car1.position[:].copy(), car0.position[:].copy()
            car0.linear_velocity[:], car1.linear_velocity[:] = \
                car1.linear_velocity[:].copy(), car0.linear_velocity[:].copy()
            car0.rotation[:], car1.rotation[:] = \
                car1.rotation[:].copy(), car0.rotation[:].copy()
            car0.angular_velocity[:], car1.angular_velocity[:] = \
                car1.angular_velocity[:].copy(), car0.angular_velocity[:].copy()
            car0.boost, car1.boost = \
                car1.boost, car0.boost

    @staticmethod
    def shuffle_players(state_wrapper: StateWrapper):
        """ The cars within a team are randomly swapped with each other """
        if len(state_wrapper.cars) <= 2:
            return
        blue_team = deepcopy(state_wrapper.blue_cars())
        orange_team = deepcopy(state_wrapper.orange_cars())
        shuffle(blue_team)
        shuffle(orange_team)

        AugmentSetter._map_cars(state_wrapper.blue_cars(), blue_team)
        AugmentSetter._map_cars(state_wrapper.orange_cars(), orange_team)

    @staticmethod
    def switch_teams(state_wrapper):
        """ Blue cars move to Orange positions, orange to blue """
        AugmentSetter._map_cars(state_wrapper.orange_cars(), state_wrapper.blue_cars())

    @staticmethod
    def mirror_front_back(state_wrapper: StateWrapper):
        AugmentSetter.switch_teams(state_wrapper)
        mul = np.array([1, -1, 1])

        for obj in [state_wrapper.ball] + state_wrapper.cars:
            obj.set_pos(*(mul * obj.position))
            obj.set_lin_vel(*(mul * obj.linear_velocity))
            obj.set_ang_vel(*(-mul * obj.angular_velocity))  # Angular velocities are negated

            if isinstance(obj, CarWrapper):
                pitch, yaw, roll = obj.rotation
                obj.set_rot(
                    pitch=pitch,
                    yaw=-yaw,
                    roll=-roll,
                )

    @staticmethod
    def mirror_left_right(state_wrapper: StateWrapper):
        mul = np.array([-1, 1, 1])

        for obj in [state_wrapper.ball] + state_wrapper.cars:
            obj.set_pos(*(mul * obj.position))
            obj.set_lin_vel(*(mul * obj.linear_velocity))
            obj.set_ang_vel(*(-mul * obj.angular_velocity))  # Angular velocities are negated

            if isinstance(obj, CarWrapper):
                pitch, yaw, roll = obj.rotation
                obj.set_rot(
                    pitch=pitch,
                    yaw=PI - yaw,
                    roll=-roll,
                )
